local M = {}
M.type = "auxiliary"
M.relevantDevice = nil
M.defaultOrder = 1000

local constants = {
    -- Conversion rates
    mpsToMph = 2.23694,
    hpToKw = 0.7457,
    torqueToHpCoef = 7120.54016,
    kWhToJ = 3600000,
    sToH = 1.0 / 3600,
    avToRPM = 9.549296596425384,
    rpmToAV = 1.0 / 9.549296596425384,
}

local min = math.min
local max = math.max
local abs = math.abs
local mathutil = require("mathutil")
local pid = require("pid")
local debugutil = require("debugutil")("hybrid", true)
local maprange = mathutil.maprange
local clamp = mathutil.clamp
local blend = mathutil.blend

local engine = nil
local gearbox = nil
local battery = nil
local esc = nil
local clutch = nil

local sounds = {
    mg1Whine = nil,
    mg2Whine = nil,
    motorRamp = nil
}

-- Calculated/initial values
local initialWheelBrakeSplit = {}

-- Quick access upvalues
local engineOn = false
local watertemp = 0
local speedMph = 0
local batteryLevel = 1
local totalPowerOutKw = 0
local throttle = 0
local throttleAggressiveness = 0
local gear = "P"

-- Other state variables
local engineDemandReason = "none"
local state = "init"
local nextState = nil
local timeSinceEngineStart = 5
local timeSinceEngineCanShutOff = 0
local timeSinceAppliedBrakes = 0
local timeSinceSwitchMG1Mode = 0
local mg1Mode = 0
local startingTime = 0
local isInChargeup = false
local lastThrottle = 0
local chargeThrottle = 0
local disableRegen = false
local desiredRegen = 0
local timeSinceAbsHit = 0
local timeSinceInit = 0
local engineLoadAdjustment = 0
local lastEngineLoadRpmError = 0
local driveMotorAdjustment = 0

local function getProportionalBatteryRatio(minLevelBias, maxLevelBias)
    minLevelBias = minLevelBias or 0
    maxLevelBias = maxLevelBias or 0

    local minPowerLevelThreshold = blend(M.minBatteryPowerLevel, M.maxBatteryPowerLevel, minLevelBias)
    local maxPowerLevelThreshold = blend(M.minBatteryPowerLevel, M.maxBatteryPowerLevel, 1 - maxLevelBias)

    return maprange(batteryLevel, minPowerLevelThreshold, maxPowerLevelThreshold)
end

local function isEngineDemandedOn()
    local inputThrottle = electrics.values.throttleOverride or input.throttle
    local maxEvOnlyThrottle

    if gear == "P" or input.brake > 0.05 or electrics.values.parkingbrake > 0 then
        maxEvOnlyThrottle = 0.02
    else
        maxEvOnlyThrottle = M.maxEvOnlyThrottle * maprange(speedMph, M.maxEvOnlySpeed / 2, M.maxEvOnlySpeed, 1.0, 0.5)
    end

    if state == "chargemode" then
        engineDemandReason = "charge mode"
        return true
    end
    if not M.enableAutoStop then
        engineDemandReason = "auto stop disabled"
        return true
    end
    if watertemp < 57 then
        engineDemandReason = "temp"
        return true
    end -- engine must be warmed up
    
    if
        gear == "R" and batteryLevel >= blend(M.minBatteryPowerLevel, M.maxBatteryPowerLevel, 0.2) and
            timeSinceEngineStart >= M.engineMinRunTime
        then
        -- After this point, if in R and battery >= 0.4, force engine off
        engineDemandReason = "none (forced off due to reverse)"
        return false
    end
    
    if speedMph > M.maxEvOnlySpeed then
        engineDemandReason = "speed"
        return true
    end -- too fast
    if obj:getDirectionVector().z > M.maxEvOnlyPitch then
        engineDemandReason = "vehicle pitch"
        return true
    end -- pitched too high up (hill start)
    if batteryLevel < M.minBatteryPowerLevel or isInChargeup then
        engineDemandReason = "battery level"
        isInChargeup = true
        return true
    end -- battery too low
    if inputThrottle > maxEvOnlyThrottle or throttleAggressiveness > M.maxEvOnlyThrottleAggressiveness then
        engineDemandReason = "throttle"
        return true
    end
    if
        engineOn and
            (not (electrics.values.parkingbrake == 1 or input.brake >= M.autoStopBrakeThreshold) or
                timeSinceAppliedBrakes < M.autoStopBrakeTime)
     then
        -- this check only applies when not in P
        if gear ~= "P" then
            engineDemandReason = "brakes"
            return true
        end
    end

    -- Easy checks are done; dynamic checks now

    -- This is the percentage of the max EV-only kW output that's available to us,
    -- based on battery level
    local batteryWorkingRatio = getProportionalBatteryRatio(0, 0.35)
    local availableEvOnlyPower = batteryWorkingRatio * M.maxEvOnlyPower
    local curPower = gearbox.motorGenerator2Power

    if engineOn then
        if curPower > 2 then
            engineDemandReason = "accelerating"
            return true
        end -- once engine is on, must coast again to shut off
    else
        if inputThrottle > 0.01 and curPower > availableEvOnlyPower then
            engineDemandReason = "accelerating"
            return true
        end
    end

    -- Put this last so it doesn't override other reasons
    if timeSinceEngineStart < M.engineMinRunTime then
        engineDemandReason = "time since start"
        return true
    end -- engine has to run for at least 4 seconds

    engineDemandReason = "none"
    return false -- by default, the engine is off
end

local function updateValues(dt)
    if not gearbox or not engine or not battery then
        return
    end
    
    local throttleAggressivenessInst = (input.throttle - lastThrottle) / dt
    
    lastThrottle = input.throttle
    engineOn = not (engine.isStalled or engine.ignitionCoef < 0.01 or electrics.values.rpm < engine.idleRPM / 2)
    watertemp = electrics.values.watertemp
    speedMph = electrics.values.wheelspeed * constants.mpsToMph
    batteryLevel = battery.remainingRatio
    gear = electrics.values.gear
    
    
    -- Rolling average
    throttleAggressiveness = throttleAggressiveness + (throttleAggressivenessInst - throttleAggressiveness) / 10
    
    local throttleRate = maprange(
        abs(throttleAggressiveness),
        M.minThrottleAggressivenessRamp,
        M.maxThrottleAggressivenessRamp,
        M.minThrottleAggressivenessFactor,
        M.maxThrottleAggressivenessFactor
    )
    
    debugutil.putValue("throttleRate", throttleRate)
    
    throttle = electrics.values.throttleOverride or M.throttleSmoother:get(input.throttle, dt * throttleRate)

    if engineOn then
        timeSinceEngineStart = timeSinceEngineStart + dt
    end

    if M.autoStopBrakeThreshold > 0 and (input.brake >= M.autoStopBrakeThreshold or electrics.values.parkingbrake == 1) and speedMph <= M.maxEvOnlySpeed then
        timeSinceAppliedBrakes = timeSinceAppliedBrakes + dt
    else
        timeSinceAppliedBrakes = 0
    end

    if isInChargeup and batteryLevel >= blend(M.minBatteryPowerLevel, M.maxBatteryPowerLevel, 0.2) then
        isInChargeup = false
    end
end

local function updateState(dt)
    local engineDemandedOn = isEngineDemandedOn()

    debugutil.putValue("Engine demanded on", tostring(engineDemandedOn))
    debugutil.putValue("Engine on reason", engineDemandReason)
    debugutil.putValue("State", state)

    -- keep things like daytime-running-lights on
    electrics.values.running = 1
    
    -- reset state variables (they can be overridden later)
    driveMotorAdjustment = 0

    -- prevent engine starter from working (since the eCVT starts it)
    engine:deactivateStarter()
    engine.starterDisabled = true

    if state == "init" then
        timeSinceInit = timeSinceInit + dt

        if timeSinceInit >= 1 then
            if engineOn then
                nextState = "running"
                engineLoadAdjustment = 0
            else
                nextState = "stopped"
            end
        end
    elseif state == "evmode" then
        engine.ignitionCoef = 0 -- Kill engine
        engine.isDisabled = true -- Prevent the user from manually restarting the engine
        electrics.values.checkengine = false -- Keep this from coming on (usually tied to engine.isDisabled)
        gearbox:setInputDisengaged(true)

        if electrics.values.evMode ~= 1 then
            if M.forceEvMode then
                electrics.values.evMode = 1
            else
                gui.message({txt = "EV mode disabled"}, 2, "vehicle.hybrid.evMode")
                gearbox:setInputDisengaged(false)
                nextState = "init"
            end
        end
    elseif state == "running" then
        engine.ignitionCoef = 1
        
        if not engineOn then
            -- engine died?

            timeSinceInit = 0
            nextState = "init"
        elseif electrics.values.evMode == 1 then
            gui.message({txt = "EV mode enabled"}, 2, "vehicle.hybrid.evMode")
            nextState = "evmode"
        elseif electrics.values.chargeMode == 1 and gear == "P" then
            gui.message({txt = "Charge mode enabled"}, 2, "vehicle.hybrid.chargeMode")
            chargeThrottle = 0
            nextState = "chargemode"
        else
            if engineDemandedOn then
                timeSinceEngineCanShutOff = 0
            else
                if gear ~= "N" then -- engine keeps state in N
                    timeSinceEngineCanShutOff = timeSinceEngineCanShutOff + dt

                    if timeSinceEngineCanShutOff > 1 then
                        engine.ignitionCoef = 0 -- Kill engine
                        engine.isDisabled = true -- Prevent the user from manually restarting the engine
                        electrics.values.checkengine = false -- Keep this from coming on (usually tied to engine.isDisabled)

                        nextState = "stopped"
                    end
                end
            end
        end
    elseif state == "chargemode" then
        local inverseRelativePowerLevel = maprange(batteryLevel, 0.85, 0.9, 1.0, 0.0)
        local maxFastChargePower = M.maxFastChargePower or gearbox.maxMG1Power
        local curGenOutput = -min(totalPowerOutKw, 0)
        local targGenOutput = inverseRelativePowerLevel * maxFastChargePower

        chargeThrottle = clamp(M.chargeThrottlePid:get(targGenOutput, curGenOutput, dt), 0, 1)

        if gear ~= "P" or electrics.values.evMode == 1 or batteryLevel >= 0.898 then
            electrics.values.chargeMode = 0
        end

        if electrics.values.chargeMode ~= 1 then
            gui.message({txt = "Charge mode disabled"}, 2, "vehicle.hybrid.chargeMode")
            nextState = "init"
        end
    elseif state == "stopped" then
        timeSinceEngineCanShutOff = 0
        engine.ignitionCoef = 0
        electrics.values.checkengine = false -- Keep this from coming on (usually tied to engine.isDisabled)
        electrics.values.iceThrottle = 0
        gearbox:setRequestedMG1TorqueCoef(0)

        if electrics.values.evMode == 1 then
            gui.message({txt = "EV mode enabled"}, 2, "vehicle.hybrid.evMode")
            nextState = "evmode"
        elseif gear ~= "N" and (engineDemandedOn or electrics.values.chargeMode == 1) then
            engine.isDisabled = false -- allow engine to restart

            if electrics.values.rpm < engine.idleRPM then
                nextState = "startup"
                timeSinceEngineStart = 0
                startingTime = 0
            else
                engine.ignitionCoef = 1 -- engine already spinning, engage ignition
                engineLoadAdjustment = 0
                nextState = "running"
            end
        end
    elseif state == "startup" then
        local targetRPM = engine.idleRPM
        local actualRPM = electrics.values.rpm
        local rpmError = targetRPM - actualRPM
        local targetStarterTorque = maprange(rpmError, 0, 500, 0, M.engineStarterTorqueCoef)
        
        engine.ignitionCoef = maprange(actualRPM, targetRPM - 300, targetRPM, 0, 1)
        electrics.values.iceThrottle = 0
        M.engineThrottleSmoother:reset()
        startingTime = startingTime + dt
        
        if engineOn then
            if timeSinceEngineStart > 0.25 then
                engine.ignitionCoef = 1
                targetStarterTorque = 0
                engineLoadAdjustment = 0
                driveMotorAdjustment = 0
                nextState = "running"
            end
        else
            if startingTime > 3 then
                -- Something's not right; retry at init
                
                targetStarterTorque = 0
                timeSinceInit = 0
                driveMotorAdjustment = 0
                nextState = "init"
            end
        end
        
        local starterTorque = M.engineStarterSmoother:get(targetStarterTorque, dt)
        local starterTorqueCompensation = starterTorque * (gearbox.ringGearTeeth / gearbox.sunGearTeeth)
        
        gearbox:setMG1PowerOverride(maprange(batteryLevel, 0.05, 0.1, 0, gearbox.maxMG1Power))
        gearbox:setMG1TorqueDirection(1)
        
        if gearbox.sunGearAV < (-gearbox.generatorMaxLoadAV * 5) then
            gearbox:setRequestedMG1TorqueCoef(-starterTorque)
        else
            gearbox:setRequestedMG1TorqueCoef(starterTorque)
        end    
        
        driveMotorAdjustment = starterTorqueCompensation
    end
end

local function updateBattery(dt)
    totalPowerOutKw = M.idlePowerDraw

    local mg1Kw = gearbox.motorGenerator1Power
    local mg2Kw = gearbox.motorGenerator2Power
    
    totalPowerOutKw = totalPowerOutKw + mg1Kw

    if mg2Kw > 0 then
        totalPowerOutKw = totalPowerOutKw + mg2Kw
    else
        totalPowerOutKw = totalPowerOutKw + (mg2Kw * M.regenEfficiency)
    end

    local totalPowerOutKj = totalPowerOutKw * dt

    debugutil.putValue("Battery power (kW)", totalPowerOutKw)
    debugutil.putValue("Generator power (kW)", mg1Kw)

    battery.storedEnergy = mathutil.clamp(battery.storedEnergy - totalPowerOutKj * 1000, 0, battery.initialStoredEnergy)
end

local function updateSounds(dt)
    -- Update sounds
    local mg1Power = abs(gearbox.motorGenerator1Power)
    local mg2Power = abs(gearbox.motorGenerator2Power)
    local mg1AV = abs(gearbox.sunGearAV)
    local mg2AV = abs(gearbox.ringGearAV)
    local mg1InverterPitch = maprange(mg1AV, M.inverterPitchMinSpeed, M.inverterPitchMaxSpeed, M.inverterMinPitch, M.inverterMaxPitch)
    local mg1InverterEffort = mg1Power / M.inverterMaxVolumeKw
    local mg1InverterVolume = M.inverterBaseVolume * clamp(mg1InverterEffort, 0.0, 1.0)
    local mg2InverterPitch = maprange(mg2AV, M.inverterPitchMinSpeed, M.inverterPitchMaxSpeed, M.inverterMinPitch, M.inverterMaxPitch)
    local mg2InverterEffort = mg2Power / M.inverterMaxVolumeKw
    local mg2InverterVolume = M.inverterBaseVolume * clamp(mg2InverterEffort, 0.0, 1.0)

    obj:setVolumePitch(sounds.mg1Whine, mg1InverterVolume, M.mg1PitchSmoother:get(mg1InverterPitch, dt))
    obj:setVolumePitch(sounds.mg2Whine, mg2InverterVolume, M.mg2PitchSmoother:get(mg2InverterPitch, dt))

    local motorPitch =
        maprange(
        speedMph,
        M.motorPitchMinSpeed,
        M.motorPitchMaxSpeed,
        M.motorMinPitch,
        M.motorMaxPitch,
        false
    )
    local motorVolume =
        maprange(
        speedMph,
        M.motorVolumeMinSpeed,
        M.motorVolumeMaxSpeed,
        0,
        M.motorBaseVolume,
        false
    )

    obj:setVolumePitch(sounds.motorRamp, motorVolume, motorPitch)
end

local function updateGUI(dt)
    local maxKw
    
    if battery and battery.jbeamData and (battery.jbeamData.maxPowerOutput or battery.jbeamData.maxPowerInput) then
        local maxOutput = battery.jbeamData.maxPowerOutput or 0
        local maxInput = battery.jbeamData.maxPowerInput or 0
        
        maxKw = max(maxOutput, maxInput)
    else
        maxKw = gearbox and max(gearbox.maxMG1Power, gearbox.maxMG2Power) or 0
    end
    
    local str = {
        soc = batteryLevel,
        kw = totalPowerOutKw,
        maxKw = maxKw,
        minSoc = state == "evmode" and 0.15 or M.minBatteryPowerLevel,
        maxSoc = M.maxBatteryPowerLevel
    }

    gui.send("hybrid", str)
    debugutil.sendStreams()
end

local function updateGFX(dt)
    updateValues(dt)

    local iceThrottle = 0.0
    
    if gearbox and engine and battery and clutch then
        local gearboxMaxMG1Torque = gearbox:getMaxMG1Torque(true)
        local gearboxMaxMG2Torque = gearbox:getMaxMG2Torque(true)
        local gearboxMG1AV = abs(gearbox.sunGearAV)
        local gearboxMG2AV = abs(gearbox.ringGearAV)
        local currentMG1MaxPower = mathutil.motorTorqueToKw(gearboxMaxMG1Torque, gearboxMG1AV)
        local currentMG2MaxPower = mathutil.motorTorqueToKw(gearboxMaxMG2Torque, gearboxMG2AV)
        local actualSpeedSign = mathutil.sign(electrics.values.avgWheelAV)
        local desiredSpeedSign = gear == "R" and -1 or 1
        local batteryWorkingRatio = 0
        
        if state == "evmode" then
            batteryWorkingRatio = maprange(batteryLevel, 0.15, M.minBatteryPowerLevel)
        else
            -- Biased a slight bit to allow some EV power at low battery
            batteryWorkingRatio = getProportionalBatteryRatio(-0.05, 0.05)
        end
        
        local availableBatteryMG1PowerCoef = maprange(batteryLevel, 0.05, 0.1)
        local availableBatteryMG2PowerCoef = maprange(batteryWorkingRatio, 0, M.relativeBatteryLevelForMaxPower)
        local maxBatteryPower = ((battery.jbeamData.maxPowerOutput or currentMG2MaxPower) - M.idlePowerDraw)
        local availableBatteryMG1Power = availableBatteryMG1PowerCoef * maxBatteryPower
        local availableBatteryMG2Power = availableBatteryMG2PowerCoef * maxBatteryPower
        local mg1GeneratedPower = max(-gearbox.motorGenerator1Power, 0)
        local mg2GeneratedPower = max(-gearbox.motorGenerator2Power, 0)
        local availableMG1Power = availableBatteryMG1Power + (mg2GeneratedPower * 0.9)
        local availableMG2Power = availableBatteryMG2Power + (mg1GeneratedPower * 0.9)
        
        gearbox:setMG1PowerOverride(M.mg1PowerSmoother:get(availableMG1Power, dt))
        gearbox:setMG2PowerOverride(M.mg2PowerSmoother:get(availableMG2Power, dt))
        
        local maxRegenPctOfMaxMG2Power = 1
        
        if battery.jbeamData.maxPowerInput then
            local gearboxMaxRegenKw = mathutil.motorTorqueToKw(gearboxMaxMG2Torque, gearboxMG2AV)
            local batteryPowerInputCoef = maprange(batteryLevel, 0.9, 0.95, 1, 0)
            local batteryPowerInputAvailable = battery.jbeamData.maxPowerInput * batteryPowerInputCoef
            
            -- Calculate what percentage of regen we can use, taking into account
            -- the maximum amount of power the battery can currently accept, minus
            -- however much power the engine is putting into the battery
            maxRegenPctOfMaxMG2Power = (batteryPowerInputAvailable) / (gearboxMaxRegenKw * M.regenEfficiency)
        end
        
        local maxRegen = disableRegen and 0 or min(M.maxRegenCoef, maxRegenPctOfMaxMG2Power)
        local throttleOffset = max(maxRegen * (gear == "L" and 0.25 or M.evCoastRegen), 0.0001)
        local posThrottle = maprange(throttle, throttleOffset, 1)
        local negThrottle = maprange(throttle, throttleOffset, 0)
        local requestedThrottle = posThrottle
        local requestedRegen = 0

        -- Fill in some debug values
        debugutil.putValue("Battery SOC % (usable)", math.floor(getProportionalBatteryRatio() * 100))
        debugutil.putValue("Throttle % EV thresh.", math.floor(throttle * 100 / M.maxEvOnlyThrottle))

        if input.brake > 0 then
            if gear == "N" or gear == "P" then
                for i = 0, wheels.wheelCount - 1, 1 do
                    local wheel = wheels.wheels[i]

                    if wheel then
                        -- Reset brake split coef
                        wheel.brakeInputSplit = initialWheelBrakeSplit[i][1]
                        wheel.brakeSplitCoef = initialWheelBrakeSplit[i][2]
                    end
                end

                debugutil.putValue("Friction % of full brakes", math.floor(input.brake * 100))
            else
                local maxBrakeTorque = 0.0
                local absActive = false

                for i = 0, wheels.wheelCount - 1, 1 do
                    local wheel = wheels.wheels[i]

                    if wheel then
                        -- Find total braking torque
                        if wheel.brakeTorque > maxBrakeTorque then
                            maxBrakeTorque = wheel.brakeTorque
                        end

                        -- Determine if ABS active
                        if wheel.absActive then
                            absActive = true
                        end
                    end
                end

                if disableRegen then
                    if not absActive then
                        timeSinceAbsHit = timeSinceAbsHit + dt

                        if timeSinceAbsHit > 1.0 then
                            disableRegen = false
                            timeSinceAbsHit = 0
                        end
                    end
                else
                    if absActive then
                        timeSinceAbsHit = timeSinceAbsHit + dt

                        if timeSinceAbsHit > 0.5 then
                            disableRegen = true
                            timeSinceAbsHit = 0
                        end
                    else
                        timeSinceAbsHit = 0
                    end
                end

                local regenBrakePedal = maprange(input.brake, 0, M.regenPercentOfBrakingForce)
                local regenAccelPedal = gear == "L" and negThrottle or 0
                local regen = maxRegen * clamp(regenBrakePedal + regenAccelPedal, 0, 1)
                
                desiredRegen = regen * maprange(speedMph, M.minRegenFadeSpeed, M.maxRegenFadeSpeed)
                requestedRegen = maprange(speedMph, 0, M.minRegenFadeSpeed) * regen

                local actualRegenCoef = requestedRegen < 0.01 and 0 or clamp((gearbox.motorGenerator2Torque * actualSpeedSign) / (-requestedRegen * gearboxMaxMG2Torque), 0, 1)
                local actualEffectiveRegen = actualRegenCoef * maxRegen * M.regenPercentOfBrakingForce
                local brakesCoveredByRegen = maprange(input.brake + (regenAccelPedal * M.regenPercentOfBrakingForce), actualEffectiveRegen, 1, 1, 0)
                local neededFrictionBrake = maprange(input.brake + (regenAccelPedal * M.regenPercentOfBrakingForce), brakesCoveredByRegen, 1)

                debugutil.putValue("Max Regen", maxRegen)
                debugutil.putValue("Regen % of full brakes", math.floor(actualEffectiveRegen * 100))
                debugutil.putValue("Regen % of full regen", math.floor(actualRegenCoef * 100))
                debugutil.putValue("Friction % of full brakes", math.floor(neededFrictionBrake * 100))

                for i = 0, wheels.wheelCount - 1, 1 do
                    local wd = wheels.wheels[i]

                    wd.brakeInputSplit = 1 - brakesCoveredByRegen

                    if wd.isPropulsed then
                        -- Propulsed wheels are affected by regenerative braking

                        wd.brakeSplitCoef = initialWheelBrakeSplit[i][2] * max(neededFrictionBrake - actualRegenCoef, 0)
                    else
                        -- Unpropulsed wheels are not

                        wd.brakeSplitCoef = initialWheelBrakeSplit[i][2] * neededFrictionBrake
                    end
                end
            end
        else
            disableRegen = false

            -- Regenerative braking when letting off accelerator (tiny bit in D, full in L)
            desiredRegen = negThrottle * maprange(speedMph, M.minRegenFadeSpeed, M.maxRegenFadeSpeed)
            requestedRegen = negThrottle * (gear == "L" and maxRegen or M.evCoastRegen) * maprange(speedMph, 0, M.minRegenFadeSpeed)

            local throttleThreshold = batteryWorkingRatio * M.evPercentOfTotalPower
        end

        local positiveSpeedCreep = maprange(speedMph, 0, 5, M.evCreepThrottle, 0)
        local hillAssistCreep = maprange(speedMph * actualSpeedSign * desiredSpeedSign, 0, -2, 0, M.maxHillAssistThrottle)
        local creep = max(positiveSpeedCreep, hillAssistCreep)

        if input.brake < 0.25 and electrics.values.parkingbrake < 1 and gear ~= "P" and gear ~= "N" then
            -- apply creep throttle
            
            requestedThrottle = min(requestedThrottle + creep, 1)
        end
        
        local mg1Dir = 1
        local mg1Coef = 0.0
        local mg2Coef = 0.0
        
        if gear == "N" then
            iceThrottle = throttle
            mg1Coef = 0
            mg2Coef = 0
        elseif state == "chargemode" then
            iceThrottle = chargeThrottle
            
            local chargeTargetRpm = M.fastChargeRpm or M.engineMaxRpm
            local desiredGenPower = maprange(electrics.values.rpm, chargeTargetRpm - M.engineLoadRpmTolerance, chargeTargetRpm + M.engineLoadRpmTolerance)
            
            mg1Coef = -desiredGenPower
            mg2Coef = 0
        else
            local maxMG2PositiveThrottle = clamp(availableMG2Power / currentMG2MaxPower, 0, 1)
            local mg1OverspeedCoef = M.throttleLimiterSmoother:get(maprange(abs(gearbox.sunGearAV), gearbox.maxMG1AV, gearbox.maxMG1AV + 300), dt)
            local normalizedThrottle = min(requestedThrottle, 1 - mg1OverspeedCoef)
            local evThrottle
            local maxEngineLoad = 0
            local engineLoad
            
            if input.brake > 0.1 or electrics.values.parkingbrake > 0 then
                evThrottle = 0
            elseif gear == "R" then
                evThrottle = normalizedThrottle
            else
                evThrottle = requestedThrottle
            end
            
            if state == "running" then
                local minRpm = max(engine.idleRPM, gearbox:getMinCarrierAV() * gearbox.inputGearRatio * constants.avToRPM)
                local maxRpm = min(M.engineMaxRpm, gearbox:getMaxCarrierAV() * gearbox.inputGearRatio * 1.2 * constants.avToRPM)
                
                if minRpm > maxRpm then
                    maxRpm = minRpm
                end
                
                -- TOO MANY CALCULATIONS, SEND HELP

                local availableEvCoef = availableBatteryMG2PowerCoef * M.evPercentOfTotalPower
                local throttleBandWidth = 1 - M.evPercentOfTotalPower
                local throttleBandOffset = (gear == "P" or gear == "N") and 0 or availableEvCoef
                local actualRpm = electrics.values.rpm
                local engineDemandCoef = maprange(requestedThrottle, throttleBandOffset, throttleBandOffset + throttleBandWidth)
                local targetRpm = maprange(engineDemandCoef, 0, 1, minRpm, gear == "R" and (minRpm + M.engineLoadRpmTolerance) or maxRpm)
                local targetIceThrottle = M.driveThrottlePid:get(targetRpm, actualRpm, dt)
                local throttleAggressivenessAdjust = engineDemandCoef * maprange(throttleAggressiveness, -M.maxThrottleAggressivenessRamp, M.maxThrottleAggressivenessRamp, -1, 1)
                
                M.driveThrottlePid:setILock(false)
                iceThrottle = clamp(targetIceThrottle + throttleAggressivenessAdjust, 0, 1)
                evThrottle = min(evThrottle, maxMG2PositiveThrottle)
                
                local maxEngineLoadBattery = maprange(batteryLevel, 0.9, 0.95, 1.0, 0.0)
                local maxEngineLoadRegen = clamp(1 - (desiredRegen * 2), 0.5, 1)
                local maxEngineLoadUnderrev = maprange(actualRpm, engine.idleRPM - 100, engine.idleRPM, 0, 1)
                local maxEngineLoadSpeedSign = (speedMph < 1 or desiredSpeedSign == actualSpeedSign) and 1 or 0
                local maxEngineLoadBrakes = (input.brake > 0.1 or electrics.values.parkingbrake > 0) and 0.5 or 1
                
                maxEngineLoad = mathutil.minall(
                    maxEngineLoadBattery, 
                    maxEngineLoadRegen, 
                    maxEngineLoadUnderrev, 
                    maxEngineLoadSpeedSign, 
                    maxEngineLoadBrakes
                )
                
                local rpmError = maprange(actualRpm, targetRpm - M.engineLoadRpmTolerance, targetRpm + M.engineLoadRpmTolerance, -1, 1)
                local proportionalEngineLoad = engineDemandCoef * maprange(rpmError, -1, 0)
                local overrevEngineLoad = maprange(rpmError, 0, 1, 0, 0.2)
                local batteryChargeLoadCoef = isInChargeup and 1 or maprange(batteryWorkingRatio, 0, 0.5, 1, 0)
                local batteryChargeLoad = batteryChargeLoadCoef * M.chargeupEngineLoad
                
                engineLoad = min(proportionalEngineLoad + overrevEngineLoad + batteryChargeLoad, maxEngineLoad)
                
                local engineDeloadAmount = maprange(batteryWorkingRatio, 0, 0.1, 0, M.maxEngineDeloadAmount)
                
                engineLoad = maprange(engineLoad, 0, 1, -engineDeloadAmount, 1)
                
                -- Since regen load from MG1 causes forward torque, we need to accomodate with negative torque
                -- from MG2 so that at zero throttle, the vehicle doesn't accelerate in a chargeup phase
                -- This torque is relative to the torque ratios of the gearbox
                local batteryChargeTorqueSun = batteryChargeLoad * gearboxMaxMG1Torque
                local batteryChargeTorqueRing = gearbox:getEquivalentRingTorqueForSun(abs(batteryChargeTorqueSun))
                local requiredMG2Accomodation = (1 - requestedThrottle) * (batteryChargeTorqueRing / gearboxMaxMG2Torque)
                
                if gear == "R" then
                    evThrottle = min(1, evThrottle + requiredMG2Accomodation)
                else
                    requestedRegen = min(1, requestedRegen + requiredMG2Accomodation)
                end
                
                debugutil.putValue("mg1Mode", mg1Mode)
                debugutil.putValue("availableEvCoef", availableEvCoef)
                debugutil.putValue("throttleBandWidth", throttleBandWidth)
                debugutil.putValue("throttleBandOffset", throttleBandOffset)
                debugutil.putValue("actualRpm", actualRpm)
                debugutil.putValue("engineDemandCoef", engineDemandCoef)
                debugutil.putValue("targetRpm", targetRpm)
                debugutil.putValue("targetIceThrottle", targetIceThrottle)
                debugutil.putValue("throttleAggressivenessAdjust", throttleAggressivenessAdjust)
                debugutil.putValue("maxEngineLoadBattery", maxEngineLoadBattery)
                debugutil.putValue("maxEngineLoadRegen", maxEngineLoadRegen)
                debugutil.putValue("maxEngineLoadUnderrev", maxEngineLoadUnderrev)
                debugutil.putValue("maxEngineLoadSpeedSign", maxEngineLoadSpeedSign)
                debugutil.putValue("maxEngineLoad", maxEngineLoad)
                debugutil.putValue("rpmError", rpmError)
                debugutil.putValue("proportionalEngineLoad", proportionalEngineLoad)
                debugutil.putValue("overrevEngineLoad", overrevEngineLoad)
                debugutil.putValue("batteryChargeLoadCoef", batteryChargeLoadCoef)
                debugutil.putValue("batteryChargeLoad", batteryChargeLoad)
                debugutil.putValue("engineLoad", engineLoad)
                debugutil.putValue("mg1MaxTorque", mg1MaxTorque)
                debugutil.putValue("mg2MaxTorque", mg2MaxTorque)
                debugutil.putValue("batteryChargeTorqueSun", batteryChargeTorqueSun)
                debugutil.putValue("batteryChargeTorqueRing", batteryChargeTorqueRing)
                debugutil.putValue("requiredMG2Accomodation", requiredMG2Accomodation)
                debugutil.putValue("evThrottle", evThrottle)
                debugutil.putValue("requestedRegen", requestedRegen)
            else
                M.driveThrottlePid:reset()
                M.driveThrottlePid:setILock(true)
                M.engineThrottleSmoother:reset()
                
                iceThrottle = 0
                engineLoad = 0
            end
            
            debugutil.putValue("Available EV Throttle", maxMg2PositiveThrottle)
            debugutil.putValue("EV Throttle", evThrottle)
            
            if gearbox.sunGearAV < gearbox.generatorMaxLoadAV then
                mg1Mode = 1
            elseif gearbox.sunGearAV > gearbox.generatorMaxLoadAV * 2 then
                mg1Mode = 0
            end
            
            if mg1Mode == 0 then
                -- engine load can be achieved using gen drag on MG1
                
                mg1Dir = 1
                mg1Coef = -engineLoad
                mg2Coef = gear == "P" and 0 or (evThrottle - requestedRegen)
            else
                -- MG1 spinning too slow (or backwards) for gen drag; use negative power instead
                -- in this case, we have to regen off of MG2 to provide power for MG1
                
                mg2Coef = gear == "P" and 0 or (evThrottle - requestedRegen - gearbox:getEquivalentSunTorqueForRing(max(engineLoad, 0)))
                mg1Dir = -sign(engineLoad)
                mg1Coef = engineLoad
            end
        end
        
        if state ~= "evmode" then
            gearbox:setInputDisengaged(gear == "N")
        end
        
        if electrics.values[clutch.electricsClutchRatioName] < 0.001 then
            -- ICE is unclutched from transmission and drag is applied to MG1 when necessary
            -- to prevent overspeed
            local mg1AVError = abs(gearbox.sunGearAV) - (gearbox.maxMG1AV - 100)
            local dragCoef = clamp(mg1AVError / 100, 0, 0.1)
            
            gearbox:setMG1TorqueDirection(1)
            gearbox:setRequestedMG1TorqueCoef(dragCoef * -sign(gearbox.sunGearAV))
        else
            gearbox:setMG1TorqueDirection(mg1Dir)
            gearbox:setRequestedMG1TorqueCoef(mg1Coef)
        end
        
        gearbox:setRequestedMG2TorqueCoef(mg2Coef + driveMotorAdjustment)
        
        updateState(dt)
        updateBattery(dt)
    end
    
    electrics.values.iceThrottle = M.engineThrottleSmoother:get(iceThrottle, dt)

    if battery then
        debugutil.putValue("Battery SOC %", tostring(battery.remainingRatio * 100))
    end

    debugutil.putValue("Input Throttle", (throttle or 0))
    debugutil.putValue("ICE Throttle", electrics.values.iceThrottle)
    debugutil.putValue("Throttle Aggr.", throttleAggressiveness)

    if nextState then
        state = nextState
    end

    debugutil.putValue("ICE Torque", engine.outputTorque1)
    debugutil.putValue("ICE HP", mathutil.motorTorqueToKw(engine.outputTorque1, engine.outputAV1) * 1.34102)
    debugutil.putValue("Charge Throttle", chargeThrottle)
    debugutil.putValue("Drive motor power (kW)", gearbox.motorGenerator2Power)
    debugutil.putValue("Motor RPM", gearbox.outputAV1 * 9.549)
    debugutil.putValue("Motor max out torque", gearbox:getMaxMG2Torque() * gearbox.gearRatio)
    debugutil.putValue("Motor out torque", gearbox.motorGenerator2Torque)
    debugutil.putValue("Gearbox gear ratio", gearbox.gearRatio)

    updateGUI(dt)
    updateSounds(dt)
end

local function init(jbeamData)
    M.jbeamData = jbeamData
    
    M.startEvMode = jbeamData.startEvMode == true
    M.forceEvMode = jbeamData.forceEvMode == true
    
    M.maxEvOnlyPower = jbeamData.maxEvOnlyPower or 30 -- kW
    M.maxEvOnlySpeed = jbeamData.maxEvOnlySpeed or 38 -- MPH
    M.maxEvOnlyThrottle = jbeamData.maxEvOnlyThrottle or 0.65 -- ratio
    M.maxEvOnlyThrottleAggressiveness = jbeamData.maxEvOnlyThrottleAggressiveness or 20
    M.maxEvOnlyPitch = jbeamData.maxEvOnlyPitch or math.pi -- 180 degrees pitch by default; will never happen (effectively disabled)
    
    M.minRegenFadeSpeed = jbeamData.minRegenFadeSpeed or 8 -- MPH
    M.maxRegenFadeSpeed = jbeamData.maxRegenFadeSpeed or 38 -- MPH
    
    M.idlePowerDraw = jbeamData.idlePowerDraw or 1 -- kW
    M.minBatteryPowerLevel = 0.25 -- ratio; low end of max-kW slope
    M.maxBatteryPowerLevel = 0.7 -- ratio; high end of max-kW slope
    M.relativeBatteryLevelForMaxPower = jbeamData.relativeBatteryLevelForMaxPower or 0.2 -- %
    M.maxFastChargePower = jbeamData.maxFastChargePower or 40 -- kW
    M.fastChargeRpm = jbeamData.fastChargeRpm
    
    M.enableAutoStop = jbeamData.enableAutoStop ~= false
    M.autoStopBrakeThreshold = jbeamData.autoStopBrakeThreshold or 0
    M.autoStopBrakeTime = jbeamData.autoStopBrakeTime or 0
    
    M.engineMinRunTime = jbeamData.engineMinRunTime or 4 -- seconds
    M.engineMaxRpm = jbeamData.engineMaxRpm
    M.engineStarterTorqueCoef = jbeamData.engineStarterTorqueCoef or 0.25
    M.chargeupEngineLoad = jbeamData.chargeupEngineLoad or 0.2
    M.engineLoadRpmTolerance = jbeamData.engineLoadRpmTolerance or 1000
    M.maxEngineDeloadAmount = jbeamData.maxEngineDeloadAmount or 0.05
    
    M.minThrottleAggressivenessRamp = jbeamData.minThrottleAggressivenessRamp or 2
    M.maxThrottleAggressivenessRamp = jbeamData.maxThrottleAggressivenessRamp or 15
    M.minThrottleAggressivenessFactor = jbeamData.minThrottleAggressivenessFactor or 1
    M.maxThrottleAggressivenessFactor = jbeamData.maxThrottleAggressivenessFactor or 10
    
    M.maxRegenCoef = jbeamData.maxRegenCoef or 0.6
    M.regenEfficiency = jbeamData.regenEfficiency or 0.85
    M.regenPercentOfBrakingForce = jbeamData.regenPercentOfBrakingForce or 0.4
    
    M.evPercentOfTotalPower = jbeamData.evPercentOfTotalPower or 0.5
    
    M.evCreepThrottle = jbeamData.evCreepThrottle or 0.18
    M.evCoastRegen = jbeamData.evCoastRegen or 0.05
    M.maxHillAssistThrottle = jbeamData.maxHillAssistThrottle or 1.0
    
    M.inverterMaxVolumeKw = jbeamData.inverterMaxVolumeKw or 50
    M.inverterBaseVolume = jbeamData.inverterBaseVolume or 0.65
    M.inverterPitchMinSpeed = jbeamData.inverterPitchMinSpeed or 1
    M.inverterPitchMaxSpeed = jbeamData.inverterPitchMaxSpeed or 1100
    M.inverterMinPitch = jbeamData.inverterMinPitch or 0.01
    M.inverterMaxPitch = jbeamData.inverterMaxPitch or 2.0
    M.motorBaseVolume = jbeamData.motorBaseVolume or 0.45
    M.motorVolumeMinSpeed = jbeamData.motorVolumeMinSpeed or 0.2
    M.motorVolumeMaxSpeed = jbeamData.motorVolumeMaxSpeed or 50
    M.motorPitchMinSpeed = jbeamData.motorPitchMinSpeed or 0
    M.motorPitchMaxSpeed = jbeamData.motorPitchMaxSpeed or 95
    M.motorMinPitch = jbeamData.motorMinPitch or 0.3
    M.motorMaxPitch = jbeamData.motorMaxPitch or 3.4
    
    M.chargeThrottlePid = pid.newPid(
        jbeamData.chargeThrottlePidPFactor or 2, 
        jbeamData.chargeThrottlePidIFactor or 5
    ):setScale(1 / 1000):clampOutput(0, 1)
    M.driveThrottlePid = pid.newPid(
        jbeamData.throttlePidPFactor or 1.2,
        jbeamData.throttlePidIFactor or 0.35,
        jbeamData.throttlePidDFactor or 0.2
    ):setScale(1 / 1000):clampOutput(0, 1)

    M.throttleSmoother = newTemporalSmoothingNonLinear(
        jbeamData.throttleSmoothingIn or 9.25, 
        jbeamData.throttleSmoothingOut or 9.25)
    M.throttleLimiterSmoother = newTemporalSmoothingNonLinear(jbeamData.throttleLimitSmoothing or 0.95)
    M.engineThrottleSmoother = newTemporalSmoothingNonLinear(jbeamData.engineThrottleSmoothing or 2)
    M.engineStarterSmoother = newTemporalSmoothingNonLinear(
        jbeamData.engineStarterSmoothingIn or 10,
        jbeamData.engineStarterSmoothingOut or 4)
    M.mg1PowerSmoother = newTemporalSmoothingNonLinear(jbeamData.motorPowerSmoothing or 10)
    M.mg2PowerSmoother = newTemporalSmoothingNonLinear(jbeamData.motorPowerSmoothing or 10)
        
    -- Sound smoothers
    M.mg1PitchSmoother = newTemporalSmoothingNonLinear(10)
    M.mg2PitchSmoother = newTemporalSmoothingNonLinear(10)

    -- Pre-fill debug table with empty strings to set up proper order
    debugutil.clearValues()
    debugutil.putValue("State", "")
    debugutil.putValue("Engine demanded on", "")
    debugutil.putValue("Engine on reason", "")
    debugutil.putValue("Battery power (kW)", "")
    debugutil.putValue("Generator power (kW)", "")
    debugutil.putValue("Drive motor power (kW)", "")
    debugutil.putValue("Inverse Relative Power Level", "")
    debugutil.putValue("Requested Generator Torque", "")
    debugutil.putValue("Battery SOC %", "")
    debugutil.putValue("Battery SOC % (usable)", "")
    debugutil.putValue("Throttle % EV thresh.", "")
    debugutil.putValue("Throttle Aggr.", "")
    debugutil.putValue("Input Throttle", "")
    debugutil.putValue("Charge Throttle", "")
    debugutil.putValue("ICE Throttle", "")
    debugutil.putValue("EV Throttle", "")
    debugutil.putValue("Available EV Throttle", "")
    debugutil.putValue("Max Regen", "")
    debugutil.putValue("Regen % of full brakes", "")
    debugutil.putValue("Regen % of full regen", "")
    debugutil.putValue("Friction % of full brakes", "")
    debugutil.putValue("Motor max out torque", "")
    debugutil.putValue("Motor out torque", "")
    debugutil.putValue("Motor % of max torque", "")
    debugutil.putValue("Motor RPM", "")
    debugutil.putValue("Gearbox gear ratio", "")
    debugutil.putValue("ICE Torque", "")
    debugutil.putValue("ICE HP", "")
end

local function reset()
    timeSinceInit = 0
    state = "init"
    nextState = nil
    engineLoadAdjustment = 0
    
    M.chargeThrottlePid:reset()
    M.driveThrottlePid:reset()

    init(M.jbeamData)
    M.initSecondStage()
end

local function initSounds()
    sounds.mg1Whine =
        obj:createSFXSource(
        "vehicles/hybrid_common/sounds/inverter.ogg",
        "AudioDefaultLoop3D",
        "mg1Whine",
        engine.engineNodeID
    )
    
    sounds.mg2Whine =
        obj:createSFXSource(
        "vehicles/hybrid_common/sounds/inverter.ogg",
        "AudioDefaultLoop3D",
        "mg2Whine",
        engine.engineNodeID
    )

    sounds.motorRamp =
        obj:createSFXSource(
        "vehicles/hybrid_common/sounds/motor_ramp.ogg",
        "AudioDefaultLoop3D",
        "motorRamp",
        engine.engineNodeID
    )
end

local function initSecondStage()
    engine = powertrain.getDevice("mainEngine")
    gearbox = powertrain.getDevice("gearbox")
    clutch = powertrain.getDevice("clutch")
    battery = energyStorage.getStorage("tractionBattery")

    if not gearbox or gearbox.type ~= "ecvtGearbox" then
        gearbox = nil
    end

    if battery then
        M.minBatteryPowerLevel = battery.jbeamData.minWorkingSoc or M.minBatteryPowerLevel
        M.maxBatteryPowerLevel = battery.jbeamData.maxWorkingSoc or M.maxBatteryPowerLevel

        if not M.maxFastChargePower then
            M.maxFastChargePower = battery.jbeamData.maxPowerInput
        end
    end

    if engine then
        if not M.jbeamData.engineMaxRpm then
            local torqueData = engine:getTorqueData()
        
            M.engineMaxRpm = torqueData.maxPowerRPM
        end
    end

    for i = 0, wheels.wheelCount - 1, 1 do
        if not initialWheelBrakeSplit[i] then
            initialWheelBrakeSplit[i] = {wheels.wheels[i].brakeInputSplit, wheels.wheels[i].brakeSplitCoef}
        end
    end

    if M.startEvMode then
        electrics.values.evMode = 1
    end
end

M.updateGFX = updateGFX
M.init = init
M.initSounds = initSounds
M.initSecondStage = initSecondStage
M.reset = reset

return M
