-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local M = {}
local mathutil = require("mathutil")
local debugutil = require("debugutil")("hybrid")
local constants = {
  rpmToAV = 0.104719755,
  avToRPM = 9.549296596425384,
  torqueToHpCoef = 7120.54016,
  torquePowerCoef = 9.5488
}

M.outputPorts = {[1] = true}
-- "viscouscoupling" below is a hack-around to allow a second gearbox to be piggybacked onto this one for insane hax
M.deviceCategories = {gearbox = true, viscouscoupling = true}
M.requiredExternalInertiaOutputs = {1}

local max = math.max
local min = math.min
local abs = math.abs
local floor = math.floor
local clamp = mathutil.clamp
local sign = mathutil.sign
local maprange = mathutil.maprange
local blend = mathutil.blend

local function updateVelocity(device, dt)
  -- ring gear is connected directly to output reduction gears
  device.ringGearAV = device.outputAV1 * device.finalGearRatio
  
  -- update planet carrier AV based on existing input AV and sun gear torque
  local adjustedParentAV = device.parent[device.parentOutputAVName] * device.inputGearRatio
  local adjustedParentTorque = device.parent[device.parentOutputTorqueName] / device.inputGearRatio - device.friction * min(max(adjustedParentAV, -1), 1)
  local carrierTorqueFromSun = device.motorGenerator1Torque * ((device.ringGearTeeth + device.sunGearTeeth) / device.sunGearTeeth)
  local carrierTorque = adjustedParentTorque + carrierTorqueFromSun
  
  device.planetCarrierAV = device.planetCarrierAV + carrierTorque * device.invCumulativeInertia * dt
  device.inputAV = device.planetCarrierAV / device.inputGearRatio
  device.parent[device.parentOutputAVName] = device.inputAV
  
  -- figure out sun gear from other two
  device.sunGearAV = ((device.sunGearTeeth + device.ringGearTeeth) * device.planetCarrierAV - (device.ringGearTeeth * device.ringGearAV)) / device.sunGearTeeth
end

local function updateTorque(device, dt)
  local reverseGearRatioCoef = device.reverseGearRatioCoef
  local oneWayTorque = device.oneWayTorqueSmoother:get(
    min(max(device.oneWayViscousCoef * device.outputAV1, -device.oneWayViscousTorque), device.oneWayViscousTorque)
  )

  device.oneWayTorqueSmoother:set(device.outputAV1 * reverseGearRatioCoef < 0 and oneWayTorque or 0)
  oneWayTorque = device.oneWayTorqueSmoother:value() * reverseGearRatioCoef
  
  device.motorGenerator1TorqueCoef = device.motorGenerator1TorqueDemand
  device.motorGenerator2TorqueCoef = device.motorGenerator2TorqueDemand
  
  -- input torque from engine
  local rawCarrierAV = device.parent[device.parentOutputAVName] * device.inputGearRatio
  local rawCarrierTorque = device.parent[device.parentOutputTorqueName] / device.inputGearRatio - device.friction * min(max(rawCarrierAV, -1), 1)
  
  -- motor torques
  local mg1AV = abs(device.sunGearAV)
  local mg2AV = abs(device.ringGearAV)
  local mg1MaxTorque = device:getMaxMG1Torque(device.motorGenerator1TorqueCoef < 0)
  local mg2MaxTorque = device:getMaxMG2Torque(device.motorGenerator2TorqueCoef < 0)
  local mg1Torque = 0
  local mg2Torque = 0
  
  if device.mode ~= "neutral" then
    if device.motorGenerator1TorqueCoef < 0 then
      local generatorCoef = maprange(abs(device.sunGearAV), 0, device.generatorMaxLoadAV)
      
      mg1Torque = -sign(device.sunGearAV) * generatorCoef * abs(device.motorGenerator1TorqueCoef) * mg1MaxTorque
    else
      mg1Torque = device.motorGenerator1TorqueCoef * mg1MaxTorque * device.motorGenerator1TorqueDirection
    end
    
    if device.motorGenerator2TorqueCoef < 0 then -- regenerative braking
      local actualMaxRegenTorque = maprange(abs(device.ringGearAV), device.regenFadeMinAV, device.regenFadeMaxAV, 0.001, mg2MaxTorque)
      
      mg2Torque = -sign(device.ringGearAV) * clamp(abs(device.motorGenerator2TorqueCoef) * mg2MaxTorque, 0, actualMaxRegenTorque)
    else
      mg2Torque = reverseGearRatioCoef * abs(device.motorGenerator2TorqueCoef * mg2MaxTorque)
    end
  end
  
  device.motorGenerator1Torque = device.mg1TorqueSmoother:get(mg1Torque, dt)
  device.motorGenerator2Torque = device.mg2TorqueSmoother:get(mg2Torque, dt)
  
  -- motor powers
  local mg1PowerCoef = device.motorGenerator1TorqueCoef < 0 and 1 or device:getPowerCoef(device.sunGearAV)
  local mg2PowerCoef = device.motorGenerator2TorqueCoef < 0 and 1 or device:getPowerCoef(device.ringGearAV)
  
  device.motorGenerator1Power = mg1PowerCoef * abs(mathutil.motorTorqueToKw(mg1Torque, mg1AV)) * sign(device.motorGenerator1TorqueCoef)
  device.motorGenerator2Power = mg2PowerCoef * abs(mathutil.motorTorqueToKw(mg2Torque, mg2AV)) * sign(device.motorGenerator2TorqueCoef)
  
  if abs(device.outputAV1) < 100 then
    device.parkEngaged = 1
  end
  
  -- output torque is the same as ring gear torque, which is a sum of
  -- the torque of MG2 plus the ratio-adjusted torques from the input
  -- and MG1
  
  local ringTorqueFromSun = device.motorGenerator1Torque * (device.ringGearTeeth / device.sunGearTeeth)
  local ringTorque = device.motorGenerator2Torque - ringTorqueFromSun
  
  device.outputTorque1 = ringTorque * device.finalGearRatio
  
  if device.mode == "park" then
    device.parkClutchAngle = min(max(device.parkClutchAngle + device.outputAV1 * dt, -device.maxParkClutchAngle), device.maxParkClutchAngle)
    device.outputTorque1 = device.outputTorque1 + -device.parkClutchAngle * device.parkLockSpring * device.parkEngaged * device.lockCoef
  end
  
  device.inputEngaged = device.mode ~= "neutral" and not device.inputForceDisengaged
  
  if not _G.gbDebugOff then
    local function formatAv(av)
      local dispRpm = floor(av * constants.avToRPM * 100) / 100
      local dispAv = floor(av * 100) / 100
      
      return dispRpm .. " (" .. dispAv .. ")"
    end
    
    debugutil.putValue("[GB] MG1 Torque", device.motorGenerator1Torque)
    debugutil.putValue("[GB] MG1 Max Torque", mg1MaxTorque)
    debugutil.putValue("[GB] MG1 Torque Demand", device.motorGenerator1TorqueDemand)
    debugutil.putValue("[GB] MG1 Torque Coef", device.motorGenerator1TorqueCoef)
    debugutil.putValue("[GB] MG1 Power", device.motorGenerator1Power)
    debugutil.putValue("[GB] MG2 Torque", device.motorGenerator2Torque)
    debugutil.putValue("[GB] MG2 Max Torque", mg2MaxTorque)
    debugutil.putValue("[GB] MG2 Torque Demand", device.motorGenerator2TorqueDemand)
    debugutil.putValue("[GB] MG2 Torque Coef", device.motorGenerator2TorqueCoef)
    debugutil.putValue("[GB] MG2 Power", device.motorGenerator2Power)
    debugutil.putValue("[GB] Sun Gear RPM (AV)", formatAv(device.sunGearAV))
    debugutil.putValue("[GB] Carrier RPM (AV)", formatAv(device.planetCarrierAV))
    debugutil.putValue("[GB] Ring Gear RPM (AV)", formatAv(device.ringGearAV))
    debugutil.putValue("[GB] Output Torque", device.outputTorque1)
  else
    debugutil.putValue("[GB] MG1 Torque", nil)
    debugutil.putValue("[GB] MG1 Max Torque", nil)
    debugutil.putValue("[GB] MG1 Torque Demand", nil)
    debugutil.putValue("[GB] MG1 Torque Coef", nil)
    debugutil.putValue("[GB] MG1 Power", nil)
    debugutil.putValue("[GB] MG2 Torque", nil)
    debugutil.putValue("[GB] MG2 Max Torque", nil)
    debugutil.putValue("[GB] MG2 Torque Demand", nil)
    debugutil.putValue("[GB] MG2 Torque Coef", nil)
    debugutil.putValue("[GB] MG2 Power", nil)
    debugutil.putValue("[GB] Sun Gear RPM (AV)", nil)
    debugutil.putValue("[GB] Carrier RPM (AV)", nil)
    debugutil.putValue("[GB] Ring Gear RPM (AV)", nil)
    debugutil.putValue("[GB] Output Torque", nil)
  end
end

local function selectUpdates(device)
  device.velocityUpdate = updateVelocity
  device.torqueUpdate = updateTorque
  device.parkEngaged = 0
end

local function validate(device)
  if not device.parent.deviceCategories.clutchlike then
    log("E", "ecvtGearbox.validate", "Parent device is not a clutch-like device...")
    log("E", "ecvtGearbox.validate", "Actual parent:")
    log("E", "ecvtGearbox.validate", powertrain.dumpsDeviceData(device.parent))
    return false
  end
  
  if not device.parent.parent or not device.parent.parent.deviceCategories.engine then
    log("E", "ecvtGearbox.validate", "Clutch parent's parent device is not an engine...")
    log("E", "ecvtGearbox.validate", "Actual clutch parent's parent:")
    log("E", "ecvtGearbox.validate", powertrain.dumpsDeviceData(device.parent.parent))
    return false
  end
  
  return true
end

local function setMode(device, mode)
  device.mode = mode
  device.reverseGearRatioCoef = mode == "reverse" and -1 or 1
  selectUpdates(device)
end

local function setGearRatio(device, ratio)
  selectUpdates(device)
end

local function setInputDisengaged(device, inputDisengaged)
  device.inputForceDisengaged = inputDisengaged
end

local function setRequestedMG1TorqueCoef(device, requestedCoef)
  device.motorGenerator1TorqueDemand = clamp(requestedCoef, -1, 1)
end

local function setRequestedMG2TorqueCoef(device, requestedCoef)
  device.motorGenerator2TorqueDemand = clamp(requestedCoef, -1, 1)
end

local function setMG1PowerOverride(device, maxPowerOverride)
  device.maxMG1PowerOverride = maxPowerOverride
end

local function setMG2PowerOverride(device, maxPowerOverride)
  device.maxMG2PowerOverride = maxPowerOverride
end

local function setMG1TorqueDirection(device, direction)
  device.motorGenerator1TorqueDirection = (direction == -1) and -1 or 1
end

local function getEquivalentSunTorqueForRing(device, ringTorque)
  return ringTorque * (device.sunGearTeeth / device.ringGearTeeth)
end

local function getEquivalentRingTorqueForSun(device, sunTorque)
  return sunTorque * (device.ringGearTeeth / device.sunGearTeeth)
end

local function setLock(device, enabled)
  device.lockCoef = enabled and 0 or 1
end

local function calculateInertia(device)
  local outputInertia = 0
  
  if device.children and #device.children > 0 then
    local child = device.children[1]
    
    outputInertia = child.cumulativeInertia
  end
  
  -- gear ratio of output to sun gear (which acts as virtual mass)
  local sunCarrierRatio = (device.ringGearTeeth + device.sunGearTeeth) / device.sunGearTeeth
  local sunCarrierInertia = device.sunGearInertia / (sunCarrierRatio * sunCarrierRatio)
  local totalInertia = device.carrierInertia + sunCarrierInertia

  device.cumulativeInertia = totalInertia / (device.inputGearRatio * device.inputGearRatio)
  device.invCumulativeInertia = 1 / device.cumulativeInertia

  device.parkLockSpring =
    device.jbeamData.parkLockSpring or (powertrain.stabilityCoef * powertrain.stabilityCoef * outputInertia * 0.5) --Nm/rad
  device.maxParkClutchAngle = device.parkLockTorque / device.parkLockSpring

  device.cumulativeGearRatio = 1
  device.maxCumulativeGearRatio = 1
end

local function getMaxMG1Torque(device, ignoreOverride)
  local mg1AV = abs(device.sunGearAV)
  local maxPower = ignoreOverride and device.maxMG1Power or (device.maxMG1PowerOverride or device.maxMG1Power)
  local maxTorque = mathutil.motorKwToTorque(maxPower, mg1AV)
  
  return min(maxTorque, device.maxMG1Torque)
end

local function getMaxMG2Torque(device, ignoreOverride)
  local mg2AV = abs(device.ringGearAV)
  local maxPower = ignoreOverride and device.maxMG2Power or (device.maxMG2PowerOverride or device.maxMG2Power)
  local maxTorque = mathutil.motorKwToTorque(maxPower, mg2AV)
  
  return min(maxTorque, device.maxMG2Torque)
end

local function getPowerCoef(device, motorAv)
  return maprange(abs(motorAv), 0, device.minPowerAV, device.startingPowerCoef, 1)
end

local function getMinCarrierAV(device)
  return (device.sunGearTeeth * -device.maxMG1AV + device.ringGearTeeth * device.ringGearAV) / (device.sunGearTeeth + device.ringGearTeeth)
end

local function getMaxCarrierAV(device)
  return (device.sunGearTeeth * device.maxMG1AV + device.ringGearTeeth * device.ringGearAV) / (device.sunGearTeeth + device.ringGearTeeth)
end

local function updateGFX(device, dt)
  if streams.willSend("ecvt") then
    gui.send("ecvt", {
      ringAv = device.ringGearAV,
      sunAv = device.sunGearAV,
      carrierAv = device.planetCarrierAV,
      ringTeeth = device.ringGearTeeth,
      sunTeeth = device.sunGearTeeth,
      planetsTeeth = device.planetGearTeeth
    })
  end
end

local function reset(device)
  local jbeamData = device.jbeamData
  device.gearRatio = jbeamData.gearRatio
  device.friction = jbeamData.friction or 0
  device.cumulativeInertia = 1
  device.cumulativeGearRatio = 1
  device.maxCumulativeGearRatio = 1

  device.outputAV1 = 0
  device.inputAV = 0
  device.outputTorque1 = 0
  device.planetCarrierAV = 0
  device.isBroken = false
  device.ringGearAV = 0

  device.lockCoef = 1
  device.reverseGearRatioCoef = 1
  device.parkClutchAngle = 0
  device.inputEngaged = true

  --one way viscous coupling (prevents rolling backwards)
  device.oneWayTorqueSmoother:reset()

  selectUpdates(device)
end

local function new(jbeamData)
  local device = {
    deviceCategories = shallowcopy(M.deviceCategories),
    requiredExternalInertiaOutputs = shallowcopy(M.requiredExternalInertiaOutputs),
    outputPorts = shallowcopy(M.outputPorts),
    name = jbeamData.name,
    type = jbeamData.type,
    inputName = jbeamData.inputName,
    inputIndex = jbeamData.inputIndex,
    gearRatio = jbeamData.gearRatio,
    finalGearRatio = jbeamData.finalGearRatio or 3.905,
    inputGearRatio = jbeamData.inputGearRatio or 1.0,
    friction = jbeamData.friction or 200,
    cumulativeInertia = 1,
    cumulativeGearRatio = 1,
    maxCumulativeGearRatio = 1,
    isPhysicallyDisconnected = true,
    
    outputAV1 = 0,
    inputAV = 0,
    sunGearAV = 0,
    planetCarrierAV = 0,
    ringGearAV = 0,
    
    outputTorque1 = 0,
    motorGenerator1TorqueDemand = 0,
    motorGenerator1Torque = 0,
    motorGenerator1TorqueCoef = 0,
    motorGenerator1TorqueDirection = 1,
    motorGenerator2TorqueDemand = 0,
    motorGenerator2Torque = 0,
    motorGenerator2TorqueCoef = 0,
    
    motorGenerator1Power = 0,
    motorGenerator2Power = 0,
    
    isBroken = false,
    lockCoef = 1,
    reverseGearRatioCoef = 1,
    gearCount = 1,
    minGearIndex = 1,
    maxGearIndex = 1,
    gearIndex = 1,
    sunGearTeeth = jbeamData.sunGearTeeth or 40,
    planetGearTeeth = jbeamData.planetGearTeeth or 23,
    ringGearTeeth = jbeamData.ringGearTeeth or 86,
    sunGearInertia = jbeamData.sunGearInertia or 0.0483,
    carrierInertia = jbeamData.carrierInertia or 0.1202,
    
    -- functions
    updateGFX = updateGFX,
    reset = reset,
    setMode = setMode,
    setRequestedMG1TorqueCoef = setRequestedMG1TorqueCoef,
    setRequestedMG2TorqueCoef = setRequestedMG2TorqueCoef,
    setMG1TorqueDirection = setMG1TorqueDirection,
    validate = validate,
    calculateInertia = calculateInertia,
    setLock = setLock,
    setGearRatio = setGearRatio,
    setInputDisengaged = setInputDisengaged,
    getMaxMG1Torque = getMaxMG1Torque,
    getMaxMG2Torque = getMaxMG2Torque,
    setMG1PowerOverride = setMG1PowerOverride,
    setMG2PowerOverride = setMG2PowerOverride,
    getPowerCoef = getPowerCoef,
    getMinCarrierAV = getMinCarrierAV,
    getMaxCarrierAV = getMaxCarrierAV,
    getEquivalentSunTorqueForRing = getEquivalentSunTorqueForRing,
    getEquivalentRingTorqueForSun = getEquivalentRingTorqueForSun,
  }

  --gearbox park locking clutch
  device.parkClutchAngle = 0
  device.parkLockTorque = jbeamData.parkLockTorque or 1000 --Nm

  --one way viscous coupling (prevents rolling backwards)
  device.oneWayViscousCoef = jbeamData.oneWayViscousCoef or 0
  device.oneWayViscousTorque = jbeamData.oneWayViscousTorque or device.oneWayViscousCoef * 25
  device.oneWayTorqueSmoother = newExponentialSmoothing(jbeamData.oneWayViscousSmoothing or 50)

  --electric stuff
  device.inputEngaged = true -- false to disengage engine (as if in neutral)
  device.inputForceDisengaged = false
  device.maxMG1Torque = jbeamData.maxMG1Torque or 240 -- Nm
  device.maxMG1AV = jbeamData.maxMG1AV or 900 -- AV (speed at which torque fades to 0)
  device.maxMG1Power = jbeamData.maxMG1Power or mathutil.motorTorqueToKw(device.maxMG1Torque, device.maxMG1AV)
  device.maxMG1PowerOverride = nil
  device.maxMG2Torque = jbeamData.maxMG2Torque or 340 -- Nm
  device.maxMG2AV = jbeamData.maxMG2AV or 1150 -- AV (speed at which torque fades to 0)
  device.maxMG2Power = jbeamData.maxMG2Power or mathutil.motorTorqueToKw(device.maxMG2Torque, device.maxMG2AV)
  device.maxMG2PowerOverride = nil
  device.minPowerAV = jbeamData.minPowerAV or 10
  device.startingPowerCoef = jbeamData.startingPowerCoef or 2.5
  device.regenFadeMinAV = jbeamData.regenFadeMinAV or 50 -- AV
  device.regenFadeMaxAV = jbeamData.regenFadeMaxAV or 650 -- AV
  device.generatorMaxLoadAV = jbeamData.generatorMaxLoadAV or 10 -- AV
  device.smoothEngageFactor = jbeamData.smoothEngageFactor or 2.5
  device.mg1TorqueSmoother = newTemporalSmoothingNonLinear(jbeamData.mg1TorqueSmoothingIn or 16.65, jbeamData.mg1TorqueSmoothingOut or 16.85)
  device.mg2TorqueSmoother = newTemporalSmoothingNonLinear(jbeamData.mg2TorqueSmoothingIn or 5.65, jbeamData.mg2TorqueSmoothingOut or 5.85)

  if jbeamData.gearboxNode_nodes and type(jbeamData.gearboxNode_nodes) == "table" then
    device.transmissionNodeID = jbeamData.gearboxNode_nodes[1]
  end

  device.breakTriggerBeam = jbeamData.breakTriggerBeam
  if device.breakTriggerBeam and device.breakTriggerBeam == "" then
    --get rid of the break beam if it's just an empty string (cancellation)
    device.breakTriggerBeam = nil
  end

  device.jbeamData = jbeamData

  selectUpdates(device)

  return device
end

M.new = new

return M
